#![allow(dead_code)]
pub mod brain {
    tonic::include_proto!("brain");
}

use anyhow::Result;
use brain::{
    brain_daemon_service_client::BrainDaemonServiceClient, NewVmConfirmation, NewVmRequest, DeletedVmUpdate,
    NodePubkey,
};
use log::{error, debug, warn, info};
use tokio::time::{sleep, Duration};
use tokio::{sync::mpsc::Receiver, sync::mpsc::Sender, task::JoinSet};
use tokio_stream::wrappers::ReceiverStream;
use tokio_stream::StreamExt;
use tonic::transport::Channel;
use lazy_static::lazy_static;
use rand::{distributions::Alphanumeric, thread_rng, Rng};

lazy_static! {
    static ref SECURE_PUBLIC_KEY: String = generate_random_string();
}

fn generate_random_string() -> String {
    let rng = thread_rng();
    rng.sample_iter(&Alphanumeric)
        .take(16) // Adjust the length as needed
        .map(char::from)
        .collect()
}

async fn listen_for_new_vm_reqs(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<NewVmRequest>,
) -> Result<()> {
    debug!("starting listen_for_new_vm_reqs");
    let node_pubkey = SECURE_PUBLIC_KEY.clone();
    let mut grpc_stream = client
        .get_new_vm_reqs(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(newvmreq) = grpc_stream.next().await {
        info!("Received new vm request: {newvmreq:?}");
        let _ = tx.send(newvmreq?).await;
    }
    debug!("listen_for_new_vm_reqs is about to exit");
    Ok(())
}

async fn send_confirmations(
    mut client: BrainDaemonServiceClient<Channel>,
    rx: Receiver<NewVmConfirmation>,
) -> Result<()> {
    debug!("starting send_confirmations stream");
    let rx_stream = ReceiverStream::new(rx);
    client.send_vm_confirmations(rx_stream).await?;
    debug!("send_confirmations is about to exit");
    Ok(())
}

async fn listen_for_deleted_vms(
    mut client: BrainDaemonServiceClient<Channel>,
    tx: Sender<DeletedVmUpdate>,
) -> Result<()> {
    debug!("starting listen_for_new_vm_reqs");
    let node_pubkey = SECURE_PUBLIC_KEY.clone();
    let mut grpc_stream = client
        .deleted_vm_updates(NodePubkey { node_pubkey })
        .await?
        .into_inner();
    while let Some(deleted_vm) = grpc_stream.next().await {
        info!("Received deleted vm: {deleted_vm:?}");
        let _ = tx.send(deleted_vm?).await;
    }
    debug!("listen_for_new_vm_reqs is about to exit");
    Ok(())
}

async fn connect_and_run() -> Result<()> {
    let client = BrainDaemonServiceClient::connect("http://[::1]:31337").await?;
    let mut streaming_tasks = JoinSet::new();

    let newvm_client = client.clone();
    let (tx, _newvm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_new_vm_reqs(newvm_client, tx));

    let confirm_client = client.clone();
    let (_confirm_tx, rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(send_confirmations(confirm_client, rx));

    let deletevms_client = client.clone();
    let (tx, _deletevm_rx) = tokio::sync::mpsc::channel(6);
    streaming_tasks.spawn(listen_for_deleted_vms(deletevms_client, tx));

    let task_output = streaming_tasks.join_next().await;
    warn!("One stream exited: {task_output:?}");
    Ok(())
}

async fn connection_wrapper() -> ! {
    loop {
        info!("Connecting to brain...");
        if let Err(e) = connect_and_run().await {
            error!("The connection broke: {e}");
        }
        sleep(Duration::from_secs(3)).await;
    }
}

#[tokio::main]
async fn main() {
    env_logger::builder()
        .filter_level(log::LevelFilter::Debug)
        .init();
    info!("Hello! My name is {}", SECURE_PUBLIC_KEY.clone());
    connection_wrapper().await;
}
